@@ -1,4 +1,4 @@
-/* C++ code produced by gperf version 3.0.4 */
+/* C++ code produced by gperf version 3.1 */
 /* Command-line: gperf -o -i 7 -C -k '1-4,6,9,$' -H keyword_hash -N check_identifier -t ./lexor_keyword.gperf  */
 
 #if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
@@ -25,7 +25,7 @@
       && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
       && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
 /* The character set is not based on ISO-646.  */
-#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gnu-gperf@gnu.org>."
+#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
 #endif
 
 #line 9 "./lexor_keyword.gperf"
@@ -52,13 +52,13 @@ struct lexor_keyword { const char*name; int mask; int tokenType; };
 class Lkwd
 {
 private:
-  static inline unsigned int keyword_hash (const char *str, unsigned int len);
+  static inline unsigned int keyword_hash (const char *str, size_t len);
 public:
-  static const struct lexor_keyword *check_identifier (const char *str, unsigned int len);
+  static const struct lexor_keyword *check_identifier (const char *str, size_t len);
 };
 
 inline unsigned int
-Lkwd::keyword_hash (register const char *str, register unsigned int len)
+Lkwd::keyword_hash (const char *str, size_t len)
 {
   static const unsigned short asso_values[] =
     {
@@ -89,37 +89,37 @@ Lkwd::keyword_hash (register const char *str, register unsigned int len)
       1547, 1547, 1547, 1547, 1547, 1547, 1547, 1547, 1547, 1547,
       1547, 1547, 1547, 1547, 1547, 1547
     };
-  register int hval = len;
+  unsigned int hval = len;
 
   switch (hval)
     {
       default:
-        hval += asso_values[(unsigned char)str[8]];
+        hval += asso_values[static_cast<unsigned char>(str[8])];
       /*FALLTHROUGH*/
       case 8:
       case 7:
       case 6:
-        hval += asso_values[(unsigned char)str[5]];
+        hval += asso_values[static_cast<unsigned char>(str[5])];
       /*FALLTHROUGH*/
       case 5:
       case 4:
-        hval += asso_values[(unsigned char)str[3]];
+        hval += asso_values[static_cast<unsigned char>(str[3])];
       /*FALLTHROUGH*/
       case 3:
-        hval += asso_values[(unsigned char)str[2]];
+        hval += asso_values[static_cast<unsigned char>(str[2])];
       /*FALLTHROUGH*/
       case 2:
-        hval += asso_values[(unsigned char)str[1]];
+        hval += asso_values[static_cast<unsigned char>(str[1])];
       /*FALLTHROUGH*/
       case 1:
-        hval += asso_values[(unsigned char)str[0]];
+        hval += asso_values[static_cast<unsigned char>(str[0])];
         break;
     }
-  return hval + asso_values[(unsigned char)str[len - 1]];
+  return hval + asso_values[static_cast<unsigned char>(str[len - 1])];
 }
 
 const struct lexor_keyword *
-Lkwd::check_identifier (register const char *str, register unsigned int len)
+Lkwd::check_identifier (const char *str, size_t len)
 {
   static const struct lexor_keyword wordlist[] =
     {
@@ -1148,11 +1148,11 @@ Lkwd::check_identifier (register const char *str, register unsigned int len)
 
   if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
     {
-      register int key = keyword_hash (str, len);
+      unsigned int key = keyword_hash (str, len);
 
-      if (key <= MAX_HASH_VALUE && key >= 0)
+      if (key <= MAX_HASH_VALUE)
         {
-          register const char *s = wordlist[key].name;
+          const char *s = wordlist[key].name;
 
           if (*str == *s && !strcmp (str + 1, s + 1))
             return &wordlist[key];
