@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001-2008 Stephen Williams (steve@icarus.com)
+ * Copyright (c) 2001-2018 Stephen Williams (steve@icarus.com)
  *
  *    This source code is free software; you can redistribute it
  *    and/or modify it in source code form under the terms of the GNU
@@ -14,18 +14,26 @@
  *
  *    You should have received a copy of the GNU General Public License
  *    along with this program; if not, write to the Free Software
- *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+# include  "config.h"
 # include  "schedule.h"
 # include  "vthread.h"
+# include  "vpi_priv.h"
+# include  "vvp_net_sig.h"
 # include  "slab.h"
+# include  "compile.h"
 # include  <new>
-# include  <signal.h>
-# include  <stdlib.h>
-# include  <assert.h>
-
-# include  <stdio.h>
+# include  <typeinfo>
+# include  <csignal>
+# include  <cstdlib>
+# include  <cassert>
+# include  <iostream>
+#ifdef CHECK_WITH_VALGRIND
+# include  "vvp_cleanup.h"
+# include  "ivl_alloc.h"
+#endif
 
 unsigned long count_assign_events = 0;
 unsigned long count_gen_events = 0;
@@ -50,16 +58,36 @@ struct event_s {
       virtual ~event_s() { }
       virtual void run_run(void) =0;
 
+	// Write something about the event to stderr
+      virtual void single_step_display(void);
+
 	// Fallback new/delete
       static void*operator new (size_t size) { return ::new char[size]; }
       static void operator delete(void*ptr)  { ::delete[]( (char*)ptr ); }
 };
 
+void event_s::single_step_display(void)
+{
+      std::cerr << "event_s: Step into event " << typeid(*this).name() << std::endl;
+}
+
 struct event_time_s {
-      event_time_s() { count_time_events += 1; }
+      event_time_s() {
+	    count_time_events += 1;
+	    start = 0;
+	    active = 0;
+	    inactive = 0;
+	    nbassign = 0;
+	    rwsync = 0;
+	    rosync = 0;
+	    del_thr = 0;
+	    next = NULL;
+      }
       vvp_time64_t delay;
 
+      struct event_s*start;
       struct event_s*active;
+      struct event_s*inactive;
       struct event_s*nbassign;
       struct event_s*rwsync;
       struct event_s*rosync;
@@ -75,12 +103,18 @@ vvp_gen_event_s::~vvp_gen_event_s()
 {
 }
 
+void vvp_gen_event_s::single_step_display(void)
+{
+      cerr << "vvp_gen_event_s: Step into event " << typeid(*this).name() << endl;
+}
+
 /*
  * Derived event types
  */
 struct vthread_event_s : public event_s {
       vthread_t thr;
       void run_run(void);
+      void single_step_display(void);
 
       static void* operator new(size_t);
       static void operator delete(void*);
@@ -92,6 +126,14 @@ void vthread_event_s::run_run(void)
       vthread_run(thr);
 }
 
+void vthread_event_s::single_step_display(void)
+{
+      __vpiScope*scope = vthread_scope(thr);
+      cerr << "vthread_event: Resume thread"
+	   << " scope=" << scope->vpi_get_str(vpiFullName)
+	   << endl;
+}
+
 static const size_t VTHR_CHUNK_COUNT = 8192 / sizeof(struct vthread_event_s);
 static slab_t<sizeof(vthread_event_s),VTHR_CHUNK_COUNT> vthread_event_heap;
 
@@ -101,14 +143,15 @@ inline void* vthread_event_s::operator new(size_t size)
       return vthread_event_heap.alloc_slab();
 }
 
-void vthread_event_s::operator delete(void*ptr)
+void vthread_event_s::operator delete(void*dptr)
 {
-      vthread_event_heap.free_slab(ptr);
+      vthread_event_heap.free_slab(dptr);
 }
 
 struct del_thr_event_s : public event_s {
       vthread_t thr;
       void run_run(void);
+      void single_step_display(void);
 };
 
 void del_thr_event_s::run_run(void)
@@ -116,12 +159,19 @@ void del_thr_event_s::run_run(void)
       vthread_delete(thr);
 }
 
+void del_thr_event_s::single_step_display(void)
+{
+      __vpiScope*scope = vthread_scope(thr);
+      cerr << "del_thr_event: Reap completed thread"
+	   << " scope=" << scope->vpi_get_str(vpiFullName) << endl;
+}
+
 struct assign_vector4_event_s  : public event_s {
 	/* The default constructor. */
-      assign_vector4_event_s(const vvp_vector4_t&that) : val(that) { }
-	/* A constructor that makes the val directly. */
-      assign_vector4_event_s(const vvp_vector4_t&that, unsigned adr, unsigned wid)
-      : val(that,adr,wid) { }
+      explicit assign_vector4_event_s(const vvp_vector4_t&that) : val(that) {
+	    base = 0;
+	    vwid = 0;
+      }
 
 	/* Where to do the assign. */
       vvp_net_ptr_t ptr;
@@ -132,6 +182,7 @@ struct assign_vector4_event_s  : public event_s {
 	/* Width of the destination vector. */
       unsigned vwid;
       void run_run(void);
+      void single_step_display(void);
 
       static void* operator new(size_t);
       static void operator delete(void*);
@@ -146,6 +197,12 @@ void assign_vector4_event_s::run_run(void)
 	    vvp_send_vec4(ptr, val, 0);
 }
 
+void assign_vector4_event_s::single_step_display(void)
+{
+      cerr << "assign_vector4_event: Propagate val=" << val
+	   << ", vwid=" << vwid << ", base=" << base << endl;
+}
+
 static const size_t ASSIGN4_CHUNK_COUNT = 524288 / sizeof(struct assign_vector4_event_s);
 static slab_t<sizeof(assign_vector4_event_s),ASSIGN4_CHUNK_COUNT> assign4_heap;
 
@@ -155,9 +212,9 @@ inline void* assign_vector4_event_s::operator new(size_t size)
       return assign4_heap.alloc_slab();
 }
 
-void assign_vector4_event_s::operator delete(void*ptr)
+void assign_vector4_event_s::operator delete(void*dptr)
 {
-      assign4_heap.free_slab(ptr);
+      assign4_heap.free_slab(dptr);
 }
 
 unsigned long count_assign4_pool(void) { return assign4_heap.pool; }
@@ -166,6 +223,7 @@ struct assign_vector8_event_s  : public event_s {
       vvp_net_ptr_t ptr;
       vvp_vector8_t val;
       void run_run(void);
+      void single_step_display(void);
 
       static void* operator new(size_t);
       static void operator delete(void*);
@@ -177,6 +235,11 @@ void assign_vector8_event_s::run_run(void)
       vvp_send_vec8(ptr, val);
 }
 
+void assign_vector8_event_s::single_step_display(void)
+{
+      cerr << "assign_vector8_event: Propagate val=" << val << endl;
+}
+
 static const size_t ASSIGN8_CHUNK_COUNT = 8192 / sizeof(struct assign_vector8_event_s);
 static slab_t<sizeof(assign_vector8_event_s),ASSIGN8_CHUNK_COUNT> assign8_heap;
 
@@ -186,9 +249,9 @@ inline void* assign_vector8_event_s::operator new(size_t size)
       return assign8_heap.alloc_slab();
 }
 
-void assign_vector8_event_s::operator delete(void*ptr)
+void assign_vector8_event_s::operator delete(void*dptr)
 {
-      assign8_heap.free_slab(ptr);
+      assign8_heap.free_slab(dptr);
 }
 
 unsigned long count_assign8_pool() { return assign8_heap.pool; }
@@ -197,6 +260,7 @@ struct assign_real_event_s  : public event_s {
       vvp_net_ptr_t ptr;
       double val;
       void run_run(void);
+      void single_step_display(void);
 
       static void* operator new(size_t);
       static void operator delete(void*);
@@ -208,6 +272,11 @@ void assign_real_event_s::run_run(void)
       vvp_send_real(ptr, val, 0);
 }
 
+void assign_real_event_s::single_step_display(void)
+{
+      cerr << "assign_real_event: Propagate val=" << val << endl;
+}
+
 static const size_t ASSIGNR_CHUNK_COUNT = 8192 / sizeof(struct assign_real_event_s);
 static slab_t<sizeof(assign_real_event_s),ASSIGNR_CHUNK_COUNT> assignr_heap;
 
@@ -217,9 +286,9 @@ inline void* assign_real_event_s::operator new (size_t size)
       return assignr_heap.alloc_slab();
 }
 
-void assign_real_event_s::operator delete(void*ptr)
+void assign_real_event_s::operator delete(void*dptr)
 {
-      assignr_heap.free_slab(ptr);
+      assignr_heap.free_slab(dptr);
 }
 
 unsigned long count_assign_real_pool(void) { return assignr_heap.pool; }
@@ -238,7 +307,7 @@ struct assign_array_word_s  : public event_s {
 void assign_array_word_s::run_run(void)
 {
       count_assign_events += 1;
-      array_set_word(mem, adr, off, val);
+      mem->set_word(adr, off, val);
 }
 
 static const size_t ARRAY_W_CHUNK_COUNT = 8192 / sizeof(struct assign_array_word_s);
@@ -257,10 +326,172 @@ void assign_array_word_s::operator delete(void*ptr)
 
 unsigned long count_assign_aword_pool(void) { return array_w_heap.pool; }
 
+struct force_vector4_event_s  : public event_s {
+	/* The default constructor. */
+      explicit force_vector4_event_s(const vvp_vector4_t&that): val(that) {
+	    net = NULL;
+	    base = 0;
+	    vwid = 0;
+      }
+	/* Where to do the force. */
+      vvp_net_t*net;
+	/* Value to force. */
+      vvp_vector4_t val;
+	/* Offset of the part into the destination. */
+      unsigned base;
+	/* Width of the destination vector. */
+      unsigned vwid;
+
+      void run_run(void);
+      void single_step_display(void);
+
+      static void* operator new(size_t);
+      static void operator delete(void*);
+};
+
+void force_vector4_event_s::run_run(void)
+{
+      count_assign_events += 1;
+
+      unsigned wid = val.size();
+      if ((base + wid) > vwid)
+	    wid = vwid - base;
+
+	// Make a mask of which bits are to be forced, 0 for unforced
+	// bits and 1 for forced bits.
+      vvp_vector2_t mask (vvp_vector2_t::FILL0, vwid);
+      for (unsigned idx = 0 ; idx < wid ; idx += 1)
+	    mask.set_bit(base+idx, 1);
+
+      vvp_vector4_t tmp (vwid, BIT4_Z);
+
+	// vvp_net_t::force_vec4 propagates all the bits of the
+	// forced vector value, regardless of the mask. This
+	// ensures the unforced bits retain their current value.
+      vvp_signal_value*sig = dynamic_cast<vvp_signal_value*>(net->fil);
+      assert(sig);
+      sig->vec4_value(tmp);
+
+      tmp.set_vec(base, val);
+      net->force_vec4(tmp, mask);
+}
+
+void force_vector4_event_s::single_step_display(void)
+{
+      cerr << "force_vector4_event: Force val=" << val
+	   << ", vwid=" << vwid << ", base=" << base << endl;
+}
+
+static const size_t FORCE4_CHUNK_COUNT = 8192 / sizeof(struct force_vector4_event_s);
+static slab_t<sizeof(force_vector4_event_s),FORCE4_CHUNK_COUNT> force4_heap;
+
+inline void* force_vector4_event_s::operator new(size_t size)
+{
+      assert(size == sizeof(force_vector4_event_s));
+      return force4_heap.alloc_slab();
+}
+
+void force_vector4_event_s::operator delete(void*dptr)
+{
+      force4_heap.free_slab(dptr);
+}
+
+unsigned long count_force4_pool(void) { return force4_heap.pool; }
+
+/*
+ * This class supports the propagation of vec4 outputs from a
+ * vvp_net_t object.
+ */
+struct propagate_vector4_event_s : public event_s {
+	/* The default constructor. */
+      explicit propagate_vector4_event_s(const vvp_vector4_t&that) : val(that) {
+	    net = NULL;
+      }
+	/* A constructor that makes the val directly. */
+      propagate_vector4_event_s(const vvp_vector4_t&that, unsigned adr, unsigned wid)
+      : val(that,adr,wid) {
+	    net = NULL;
+      }
+
+	/* Propagate the output of this net. */
+      vvp_net_t*net;
+	/* value to propagate */
+      vvp_vector4_t val;
+	/* Action */
+      void run_run(void);
+      void single_step_display(void);
+};
+
+void propagate_vector4_event_s::run_run(void)
+{
+      net->send_vec4(val, 0);
+}
+
+void propagate_vector4_event_s::single_step_display(void)
+{
+      cerr << "propagate_vector4_event: Propagate val=" << val << endl;
+}
+
+/*
+ * This class supports the propagation of real outputs from a
+ * vvp_net_t object.
+ */
+struct propagate_real_event_s : public event_s {
+	/* Propagate the output of this net. */
+      vvp_net_t*net;
+	/* value to propagate */
+      double val;
+	/* Action */
+      void run_run(void);
+      void single_step_display(void);
+};
+
+void propagate_real_event_s::run_run(void)
+{
+      net->send_real(val, 0);
+}
+
+void propagate_real_event_s::single_step_display(void)
+{
+      cerr << "propagate_real_event: Propagate val=" << val << endl;
+}
+
+struct assign_array_r_word_s  : public event_s {
+      vvp_array_t mem;
+      unsigned adr;
+      double val;
+      void run_run(void);
+
+      static void* operator new(size_t);
+      static void operator delete(void*);
+};
+
+void assign_array_r_word_s::run_run(void)
+{
+      count_assign_events += 1;
+      mem->set_word(adr, val);
+}
+static const size_t ARRAY_R_W_CHUNK_COUNT = 8192 / sizeof(struct assign_array_r_word_s);
+static slab_t<sizeof(assign_array_r_word_s),ARRAY_R_W_CHUNK_COUNT> array_r_w_heap;
+
+inline void* assign_array_r_word_s::operator new(size_t size)
+{
+      assert(size == sizeof(assign_array_r_word_s));
+      return array_r_w_heap.alloc_slab();
+}
+
+void assign_array_r_word_s::operator delete(void*ptr)
+{
+      array_r_w_heap.free_slab(ptr);
+}
+
+unsigned long count_assign_arword_pool(void) { return array_r_w_heap.pool; }
+
 struct generic_event_s : public event_s {
       vvp_gen_event_t obj;
       bool delete_obj_when_done;
       void run_run(void);
+      void single_step_display(void);
 
       static void* operator new(size_t);
       static void operator delete(void*);
@@ -276,6 +507,11 @@ void generic_event_s::run_run(void)
       }
 }
 
+void generic_event_s::single_step_display(void)
+{
+      obj->single_step_display();
+}
+
 static const size_t GENERIC_CHUNK_COUNT = 131072 / sizeof(struct generic_event_s);
 static slab_t<sizeof(generic_event_s),GENERIC_CHUNK_COUNT> generic_event_heap;
 
@@ -308,7 +544,7 @@ inline void* event_time_s::operator new (size_t size)
       return ptr;
 }
 
-inline void event_time_s::operator delete(void*ptr, size_t size)
+inline void event_time_s::operator delete(void*ptr, size_t)
 {
       event_time_heap.free_slab(ptr);
 }
@@ -330,12 +566,18 @@ static struct event_time_s* sched_list = 0;
  */
 static struct event_s* schedule_init_list = 0;
 
+/*
+ * This is the head of the list of final events.
+ */
+static struct event_s* schedule_final_list = 0;
+
 /*
  * This flag is true until a VPI task or function finishes the
  * simulation.
  */
 static bool schedule_runnable = true;
 static bool schedule_stopped_flag  = false;
+static bool schedule_single_step_flag = false;
 
 void schedule_finish(int)
 {
@@ -347,6 +589,11 @@ void schedule_stop(int)
       schedule_stopped_flag = true;
 }
 
+void schedule_single_step(int)
+{
+      schedule_single_step_flag = true;
+}
+
 bool schedule_finished(void)
 {
       return !schedule_runnable;
@@ -359,23 +606,69 @@ bool schedule_stopped(void)
 
 /*
  * These are the signal handling infrastructure. The SIGINT signal
- * leads to an implicit $stop.
+ * leads to an implicit $stop. The SIGHUP and SIGTERM signals lead
+ * to an implicit $finish.
  */
-static void signals_handler(int)
+extern bool stop_is_finish;
+
+extern "C" void signals_handler(int signum)
 {
+#ifdef __MINGW32__
+	// Windows implements the original UNIX semantics for signal,
+	// so we have to re-establish the signal handler each time a
+	// signal is caught.
+      signal(signum, &signals_handler);
+#endif
+      if (signum != SIGINT)
+	    stop_is_finish = true;
       schedule_stopped_flag = true;
 }
 
 static void signals_capture(void)
 {
-      signal(SIGINT, &signals_handler);
+#ifndef __MINGW32__
+      signal(SIGHUP,  &signals_handler);
+#endif
+      signal(SIGINT,  &signals_handler);
+      signal(SIGTERM, &signals_handler);
 }
 
 static void signals_revert(void)
 {
-      signal(SIGINT, SIG_DFL);
+#ifndef __MINGW32__
+      signal(SIGHUP,  SIG_DFL);
+#endif
+      signal(SIGINT,  SIG_DFL);
+      signal(SIGTERM, SIG_DFL);
 }
 
+/*
+ * This function puts an event on the end of the pre-simulation event queue.
+ */
+static void schedule_init_event(struct event_s*cur)
+{
+      if (schedule_init_list == 0) {
+            cur->next = cur;
+      } else {
+            cur->next = schedule_init_list->next;
+            schedule_init_list->next = cur;
+      }
+      schedule_init_list = cur;
+}
+
+/*
+ * This function puts an event on the end of the post-simulation event queue.
+ */
+static void schedule_final_event(struct event_s*cur)
+{
+      if (schedule_final_list == 0) {
+            cur->next = cur;
+      } else {
+            cur->next = schedule_final_list->next;
+            schedule_final_list->next = cur;
+      }
+      schedule_final_list = cur;
+}
 
 /*
  * This function does all the hard work of putting an event into the
@@ -383,25 +676,19 @@ static void signals_revert(void)
  * itself, and the structure is placed in the right place in the
  * queue.
  */
-typedef enum event_queue_e { SEQ_ACTIVE, SEQ_NBASSIGN, SEQ_RWSYNC, SEQ_ROSYNC,
-                             DEL_THREAD } event_queue_t;
+typedef enum event_queue_e { SEQ_START, SEQ_ACTIVE, SEQ_INACTIVE, SEQ_NBASSIGN,
+			     SEQ_RWSYNC, SEQ_ROSYNC, DEL_THREAD } event_queue_t;
 
 static void schedule_event_(struct event_s*cur, vvp_time64_t delay,
 			    event_queue_t select_queue)
 {
       cur->next = cur;
-
       struct event_time_s*ctim = sched_list;
 
       if (sched_list == 0) {
 	      /* Is the event_time list completely empty? Create the
 		 first event_time object. */
 	    ctim = new struct event_time_s;
-	    ctim->active = 0;
-	    ctim->nbassign = 0;
-	    ctim->rwsync = 0;
-	    ctim->rosync = 0;
-	    ctim->del_thr = 0;
 	    ctim->delay = delay;
 	    ctim->next  = 0;
 	    sched_list = ctim;
@@ -411,11 +698,6 @@ static void schedule_event_(struct event_s*cur, vvp_time64_t delay,
 	      /* Am I looking for an event before the first event_time?
 		 If so, create a new event_time to go in front. */
 	    struct event_time_s*tmp = new struct event_time_s;
-	    tmp->active = 0;
-	    tmp->nbassign = 0;
-	    tmp->rwsync = 0;
-	    tmp->rosync = 0;
-	    tmp->del_thr = 0;
 	    tmp->delay = delay;
 	    tmp->next = ctim;
 	    ctim->delay -= delay;
@@ -433,11 +715,6 @@ static void schedule_event_(struct event_s*cur, vvp_time64_t delay,
 
 	    if (ctim->delay > delay) {
 		  struct event_time_s*tmp = new struct event_time_s;
-		  tmp->active = 0;
-		  tmp->nbassign = 0;
-		  tmp->rwsync = 0;
-		  tmp->rosync = 0;
-		  tmp->del_thr = 0;
 		  tmp->delay = delay;
 		  tmp->next  = prev->next;
 		  prev->next = tmp;
@@ -450,11 +727,6 @@ static void schedule_event_(struct event_s*cur, vvp_time64_t delay,
 	    } else {
 		  assert(ctim->next == 0);
 		  struct event_time_s*tmp = new struct event_time_s;
-		  tmp->active = 0;
-		  tmp->nbassign = 0;
-		  tmp->rwsync = 0;
-		  tmp->rosync = 0;
-		  tmp->del_thr = 0;
 		  tmp->delay = delay - ctim->delay;
 		  tmp->next = 0;
 		  ctim->next = tmp;
@@ -467,68 +739,47 @@ static void schedule_event_(struct event_s*cur, vvp_time64_t delay,
 	   receive the event at hand. Put the event in to the
 	   appropriate list for the kind of assign we have at hand. */
 
+      struct event_s** q = 0;
+
       switch (select_queue) {
+	  case SEQ_START:
+	    q = &ctim->start;
+	    break;
 
 	  case SEQ_ACTIVE:
-	    if (ctim->active == 0) {
-		  ctim->active = cur;
+	    q = &ctim->active;
+	    break;
 
-	    } else {
-		    /* Put the cur event on the end of the active list. */
-		  cur->next = ctim->active->next;
-		  ctim->active->next = cur;
-		  ctim->active = cur;
-	    }
+	  case SEQ_INACTIVE:
+	    assert(delay == 0);
+	    q = &ctim->inactive;
 	    break;
 
 	  case SEQ_NBASSIGN:
-	    if (ctim->nbassign == 0) {
-		  ctim->nbassign = cur;
-
-	    } else {
-		    /* Put the cur event on the end of the active list. */
-		  cur->next = ctim->nbassign->next;
-		  ctim->nbassign->next = cur;
-		  ctim->nbassign = cur;
-	    }
+	    q = &ctim->nbassign;
 	    break;
 
 	  case SEQ_RWSYNC:
-	    if (ctim->rwsync == 0) {
-		  ctim->rwsync = cur;
-
-	    } else {
-		    /* Put the cur event on the end of the active list. */
-		  cur->next = ctim->rwsync->next;
-		  ctim->rwsync->next = cur;
-		  ctim->rwsync = cur;
-	    }
+	    q = &ctim->rwsync;
 	    break;
 
 	  case SEQ_ROSYNC:
-	    if (ctim->rosync == 0) {
-		  ctim->rosync = cur;
-
-	    } else {
-		    /* Put the cur event on the end of the active list. */
-		  cur->next = ctim->rosync->next;
-		  ctim->rosync->next = cur;
-		  ctim->rosync = cur;
-	    }
+	    q = &ctim->rosync;
 	    break;
 
 	  case DEL_THREAD:
-	    if (ctim->del_thr == 0) {
-		  ctim->del_thr = cur;
-
-	    } else {
-		    /* Put the cur event on the end of the active list. */
-		  cur->next = ctim->del_thr->next;
-		  ctim->del_thr->next = cur;
-		  ctim->del_thr = cur;
-	    }
+	    q = &ctim->del_thr;
 	    break;
       }
+
+      if (q) {
+	    if (*q) {
+		  /* Put the cur event on the end of the queue. */
+		  cur->next = (*q)->next;
+		  (*q)->next = cur;
+	    }
+	    *q = cur;
+      }
 }
 
 static void schedule_event_push_(struct event_s*cur)
@@ -550,7 +801,6 @@ static void schedule_event_push_(struct event_s*cur)
       ctim->active->next = cur;
 }
 
-
 void schedule_vthread(vthread_t thr, vvp_time64_t delay, bool push_flag)
 {
       struct vthread_event_s*cur = new vthread_event_s;
@@ -570,6 +820,43 @@ void schedule_vthread(vthread_t thr, vvp_time64_t delay, bool push_flag)
       }
 }
 
+void schedule_t0_trigger(vvp_net_ptr_t ptr)
+{
+      vvp_vector4_t bit (1, BIT4_X);
+      struct assign_vector4_event_s*cur = new struct assign_vector4_event_s(bit);
+      cur->ptr = ptr;
+      schedule_event_(cur, 0, SEQ_INACTIVE);
+}
+
+void schedule_inactive(vthread_t thr)
+{
+      struct vthread_event_s*cur = new vthread_event_s;
+
+      cur->thr = thr;
+      vthread_mark_scheduled(thr);
+      schedule_event_(cur, 0, SEQ_INACTIVE);
+}
+
+void schedule_init_vthread(vthread_t thr)
+{
+      struct vthread_event_s*cur = new vthread_event_s;
+
+      cur->thr = thr;
+      vthread_mark_scheduled(thr);
+
+      schedule_init_event(cur);
+}
+
+void schedule_final_vthread(vthread_t thr)
+{
+      struct vthread_event_s*cur = new vthread_event_s;
+
+      cur->thr = thr;
+      vthread_mark_scheduled(thr);
+
+      schedule_final_event(cur);
+}
+
 void schedule_assign_vector(vvp_net_ptr_t ptr,
 			    unsigned base, unsigned vwid,
 			    const vvp_vector4_t&bit,
@@ -582,16 +869,25 @@ void schedule_assign_vector(vvp_net_ptr_t ptr,
       schedule_event_(cur, delay, SEQ_NBASSIGN);
 }
 
-void schedule_assign_plucked_vector(vvp_net_ptr_t ptr,
-				    vvp_time64_t delay,
-				    const vvp_vector4_t&src,
-				    unsigned adr, unsigned wid)
+void schedule_force_vector(vvp_net_t*net,
+			    unsigned base, unsigned vwid,
+			    const vvp_vector4_t&bit,
+			    vvp_time64_t delay)
 {
-      struct assign_vector4_event_s*cur
-	    = new struct assign_vector4_event_s(src,adr,wid);
-      cur->ptr = ptr;
-      cur->vwid = 0;
-      cur->base = 0;
+      struct force_vector4_event_s*cur = new struct force_vector4_event_s(bit);
+      cur->net = net;
+      cur->base = base;
+      cur->vwid = vwid;
+      schedule_event_(cur, delay, SEQ_NBASSIGN);
+}
+
+void schedule_propagate_vector(vvp_net_t*net,
+			       vvp_time64_t delay,
+			       const vvp_vector4_t&src)
+{
+      struct propagate_vector4_event_s*cur
+	    = new struct propagate_vector4_event_s(src);
+      cur->net = net;
       schedule_event_(cur, delay, SEQ_NBASSIGN);
 }
 
@@ -609,6 +905,18 @@ void schedule_assign_array_word(vvp_array_t mem,
       schedule_event_(cur, delay, SEQ_NBASSIGN);
 }
 
+void schedule_assign_array_word(vvp_array_t mem,
+				unsigned word_addr,
+				double val,
+				vvp_time64_t delay)
+{
+      struct assign_array_r_word_s*cur = new struct assign_array_r_word_s;
+      cur->mem = mem;
+      cur->adr = word_addr;
+      cur->val = val;
+      schedule_event_(cur, delay, SEQ_NBASSIGN);
+}
+
 void schedule_set_vector(vvp_net_ptr_t ptr, const vvp_vector4_t&bit)
 {
       struct assign_vector4_event_s*cur = new struct assign_vector4_event_s(bit);
@@ -640,8 +948,15 @@ void schedule_init_vector(vvp_net_ptr_t ptr, vvp_vector4_t bit)
       cur->ptr = ptr;
       cur->base = 0;
       cur->vwid = 0;
-      cur->next = schedule_init_list;
-      schedule_init_list = cur;
+      schedule_init_event(cur);
+}
+
+void schedule_init_vector(vvp_net_ptr_t ptr, vvp_vector8_t bit)
+{
+      struct assign_vector8_event_s*cur = new struct assign_vector8_event_s;
+      cur->ptr = ptr;
+      cur->val = bit;
+      schedule_init_event(cur);
 }
 
 void schedule_init_vector(vvp_net_ptr_t ptr, double bit)
@@ -649,8 +964,22 @@ void schedule_init_vector(vvp_net_ptr_t ptr, double bit)
       struct assign_real_event_s*cur = new struct assign_real_event_s;
       cur->ptr = ptr;
       cur->val = bit;
-      cur->next = schedule_init_list;
-      schedule_init_list = cur;
+      schedule_init_event(cur);
+}
+
+void schedule_init_propagate(vvp_net_t*net, vvp_vector4_t bit)
+{
+      struct propagate_vector4_event_s*cur = new struct propagate_vector4_event_s(bit);
+      cur->net = net;
+      schedule_init_event(cur);
+}
+
+void schedule_init_propagate(vvp_net_t*net, double bit)
+{
+      struct propagate_real_event_s*cur = new struct propagate_real_event_s;
+      cur->net = net;
+      cur->val = bit;
+      schedule_init_event(cur);
 }
 
 void schedule_del_thr(vthread_t thr)
@@ -671,6 +1000,47 @@ void schedule_generic(vvp_gen_event_t obj, vvp_time64_t delay,
       cur->delete_obj_when_done = delete_when_done;
       schedule_event_(cur, delay,
 		      sync_flag? (ro_flag?SEQ_ROSYNC:SEQ_RWSYNC) : SEQ_ACTIVE);
+
+      if (sync_flag)
+	    vthread_delay_delete();
+}
+
+static bool sim_started;
+
+void schedule_functor(vvp_gen_event_t obj)
+{
+      struct generic_event_s*cur = new generic_event_s;
+
+      cur->obj = obj;
+      cur->delete_obj_when_done = false;
+      if (!sim_started) {
+            schedule_init_event(cur);
+      } else {
+            schedule_event_(cur, 0, SEQ_ACTIVE);
+      }
+}
+
+void schedule_at_start_of_simtime(vvp_gen_event_t obj, vvp_time64_t delay)
+{
+      struct generic_event_s*cur = new generic_event_s;
+
+      cur->obj = obj;
+      cur->delete_obj_when_done = false;
+      schedule_event_(cur, delay, SEQ_START);
+}
+
+/*
+ * In the vvp runtime of Icarus Verilog, the SEQ_RWSYNC time step is
+ * after all of the non-blocking assignments, so is effectively the
+ * same as the ReadWriteSync time.
+ */
+void schedule_at_end_of_simtime(vvp_gen_event_t obj, vvp_time64_t delay)
+{
+      struct generic_event_s*cur = new generic_event_s;
+
+      cur->obj = obj;
+      cur->delete_obj_when_done = false;
+      schedule_event_(cur, delay, SEQ_RWSYNC);
 }
 
 static vvp_time64_t schedule_time;
@@ -682,6 +1052,10 @@ extern void vpiStartOfSim();
 extern void vpiPostsim();
 extern void vpiNextSimTime(void);
 
+static bool sim_at_rosync = false;
+bool schedule_at_rosync(void)
+{ return sim_at_rosync; }
+
 /*
  * The scheduler uses this function to drain the rosync events of the
  * current time. The ctim object is still in the event queue, because
@@ -694,6 +1068,7 @@ extern void vpiNextSimTime(void);
  */
 static void run_rosync(struct event_time_s*ctim)
 {
+      sim_at_rosync = true;
       while (ctim->rosync) {
 	    struct event_s*cur = ctim->rosync->next;
 	    if (cur->next == cur) {
@@ -705,6 +1080,7 @@ static void run_rosync(struct event_time_s*ctim)
 	    cur->run_run();
 	    delete cur;
       }
+      sim_at_rosync = false;
 
       while (ctim->del_thr) {
 	    struct event_s*cur = ctim->del_thr->next;
@@ -718,33 +1094,61 @@ static void run_rosync(struct event_time_s*ctim)
 	    delete cur;
       }
 
-      if (ctim->active || ctim->nbassign || ctim->rwsync) {
-	    fprintf(stderr, "SCHEDULER ERROR: read-only sync events "
-		    "created RW events!\n");
+      if (ctim->active || ctim->inactive || ctim->nbassign || ctim->rwsync) {
+	    cerr << "SCHEDULER ERROR: read-only sync events "
+		 << "created RW events!" << endl;
       }
 }
 
 void schedule_simulate(void)
 {
+      bool run_finals;
+      sim_started = false;
+
       schedule_time = 0;
 
+      if (verbose_flag) {
+	    vpi_mcd_printf(1, " ...execute EndOfCompile callbacks\n");
+      }
+
       // Execute end of compile callbacks
       vpiEndOfCompile();
 
+      if (verbose_flag) {
+	    vpi_mcd_printf(1, " ...propagate initialization events\n");
+      }
+
 	// Execute initialization events.
       while (schedule_init_list) {
-	    struct event_s*cur = schedule_init_list;
-	    schedule_init_list = cur->next;
-
+	    struct event_s*cur = schedule_init_list->next;
+	    if (cur->next == cur) {
+		  schedule_init_list = 0;
+	    } else {
+		  schedule_init_list->next = cur->next;
+	    }
 	    cur->run_run();
 	    delete cur;
       }
 
+      if (verbose_flag) {
+	    vpi_mcd_printf(1, " ...execute StartOfSim callbacks\n");
+      }
+
       // Execute start of simulation callbacks
       vpiStartOfSim();
 
+      sim_started = true;
+
       signals_capture();
 
+      if (verbose_flag) {
+	    vpi_mcd_printf(1, " ...run scheduler\n");
+      }
+
+      // If there were no compiletf, etc. errors then we are going to
+      // process events and when done run the final blocks.
+      run_finals = schedule_runnable;
+
       if (schedule_runnable) while (sched_list) {
 
 	    if (schedule_stopped_flag) {
@@ -764,9 +1168,26 @@ void schedule_simulate(void)
 
 		  if (!schedule_runnable) break;
 		  schedule_time += ctim->delay;
+		    /* When the design is being traced (we are emitting
+		     * file/line information) also print any time changes. */
+		  if (show_file_line) {
+			cerr << "Advancing to simulation time: "
+			     << schedule_time << endl;
+		  }
 		  ctim->delay = 0;
 
 		  vpiNextSimTime();
+		    // Process the cbAtStartOfSimTime callbacks.
+		  while (ctim->start) {
+			struct event_s*cur = ctim->start->next;
+			if (cur->next == cur) {
+			      ctim->start = 0;
+			} else {
+			      ctim->start->next = cur->next;
+			}
+			cur->run_run();
+			delete (cur);
+		  }
 	    }
 
 
@@ -774,21 +1195,26 @@ void schedule_simulate(void)
 		 queues. If there are not events at all, then release
 		 the event_time object. */
 	    if (ctim->active == 0) {
-		  ctim->active = ctim->nbassign;
-		  ctim->nbassign = 0;
+		  ctim->active = ctim->inactive;
+		  ctim->inactive = 0;
 
 		  if (ctim->active == 0) {
-			ctim->active = ctim->rwsync;
-			ctim->rwsync = 0;
+			ctim->active = ctim->nbassign;
+			ctim->nbassign = 0;
 
-			  /* If out of rw events, then run the rosync
-			     events and delete this time step. This also
-			     deletes threads as needed. */
 			if (ctim->active == 0) {
-			      run_rosync(ctim);
-			      sched_list = ctim->next;
-			      delete ctim;
-			      continue;
+			      ctim->active = ctim->rwsync;
+			      ctim->rwsync = 0;
+
+				/* If out of rw events, then run the rosync
+				   events and delete this time step. This also
+				   deletes threads as needed. */
+			      if (ctim->active == 0) {
+				    run_rosync(ctim);
+				    sched_list = ctim->next;
+				    delete ctim;
+				    continue;
+			      }
 			}
 		  }
 	    }
@@ -803,14 +1229,53 @@ void schedule_simulate(void)
 		  ctim->active->next = cur->next;
 	    }
 
+	    if (schedule_single_step_flag) {
+		  cur->single_step_display();
+		  schedule_stopped_flag = true;
+		  schedule_single_step_flag = false;
+	    }
+
 	    cur->run_run();
 
 	    delete (cur);
       }
 
+	// Execute final events.
+      schedule_runnable = run_finals;
+      while (schedule_runnable && schedule_final_list) {
+	    struct event_s*cur = schedule_final_list->next;
+	    if (cur->next == cur) {
+		  schedule_final_list = 0;
+	    } else {
+		  schedule_final_list->next = cur->next;
+	    }
+	    cur->run_run();
+	    delete cur;
+      }
 
       signals_revert();
 
+      if (verbose_flag) {
+	    vpi_mcd_printf(1, " ...execute Postsim callbacks\n");
+      }
+
       // Execute post-simulation callbacks
       vpiPostsim();
+#ifdef CHECK_WITH_VALGRIND
+      schedule_delete();
+#endif
 }
+
+#ifdef CHECK_WITH_VALGRIND
+void schedule_delete(void)
+{
+      vthread_event_heap.delete_pool();
+      assign4_heap.delete_pool();
+      assign8_heap.delete_pool();
+      assignr_heap.delete_pool();
+      array_w_heap.delete_pool();
+      array_r_w_heap.delete_pool();
+      generic_event_heap.delete_pool();
+      event_time_heap.delete_pool();
+}
+#endif
